#define SQ1_CHANNEL 0
#define SQ2_CHANNEL 1
#define TRI_CHANNEL 2
#define NOISE_CHANNEL 3
#define DMC_CHANNEL 4

#define DELTA_INIT 0x3f

// $B%<%m%Z!<%8$rMxMQ(B
extern unsigned short SoundCounter[5];
extern unsigned short SoundIndex[5];
extern unsigned char RepeatCounter[5];
extern unsigned char Ctrl0Number[5];
extern unsigned char Ctrl1Number[5];
extern unsigned short FreqNumber[5];

#pragma zpsym ("SoundCounter");
#pragma zpsym ("SoundIndex");
#pragma zpsym ("RepeatCounter");
#pragma zpsym ("Ctrl0Number");
#pragma zpsym ("Ctrl1Number");
#pragma zpsym ("FreqNumber");

extern const unsigned  char *SoundAddr;

const unsigned short FreqTable[] = {
	0x07f2, // a
	0x0780,
	0x0714,
	0x06ae,
	0x064e,
	0x05f4,
	0x059e,
	0x054e,
	0x0501,
	0x04b9,
	0x0476,
	0x0436
};

// $B%5%&%s%I%5%&%s%I=i4|2=(B
void soundInit(void)
{
	unsigned char i;

	for (i = 0; i < 5; i++) {
		SoundIndex[i] = 0;
		SoundCounter[i] = 1;
		RepeatCounter[i] = 0;
		Ctrl0Number[i] = 0;
		Ctrl1Number[i] = 0;
		FreqNumber[i] = 0;
	}
}

// $B%5%&%s%I:F@8%W%m%7!<%8%c(B
void soundProc(unsigned char ch)
{

	// SoundCounter[ch]-- == 0 $B$G$OBLL\$i$7$$!#(B
	// $B%3%s%Q%$%i$N%P%0$+;EMM$+!)(B
	SoundCounter[ch]--;
	while (SoundCounter[ch] == 0) {
		unsigned char c;
		c = SoundAddr[SoundIndex[ch]];
		SoundIndex[ch]++;
		// $B%G!<%?KvC<(B
		if (c ==  0xff) {
			SoundIndex[ch] = 0;
			SoundCounter[ch] = 0;
			RepeatCounter[ch] = 0;
			// $B%j%T!<%H(B	
		} else if (c ==  0xfe) {
			unsigned char l,h;
			if (RepeatCounter[ch] == 0) {
				RepeatCounter[ch] = SoundAddr[SoundIndex[ch]] - 1;
			} else {
				// $B%j%T!<%HCf$OFI$_Ht$P$9(B
				RepeatCounter[ch]--;
			}
			SoundIndex[ch]++; 

			// $B3Z6J$N0LCV$r99?7(B
			l  = SoundAddr[SoundIndex[ch]];
			SoundIndex[ch]++;
			h  = SoundAddr[SoundIndex[ch]];
			SoundIndex[ch] = 0;
			SoundIndex[ch] = h << 8;
			SoundIndex[ch] += l;

			// $B%j%T!<%HJ,4t%8%c%s%W(B
		} else if (c == 0xfd) {
			if (RepeatCounter[ch] == 1) {
				unsigned char l,h;
				l  = SoundAddr[SoundIndex[ch]];
				SoundIndex[ch]++;
				h  = SoundAddr[SoundIndex[ch]];
				SoundIndex[ch] = 0;
				SoundIndex[ch] = h << 8;
				SoundIndex[ch] += l;
			} else {
				SoundIndex[ch] += 2;
			}
		} else if (c == 0xa0) {
			Ctrl0Number[ch] = SoundAddr[SoundIndex[ch]];
			SoundIndex[ch]++;
		} else if (c == 0xa1) {
			Ctrl1Number[ch] = SoundAddr[SoundIndex[ch]];
			SoundIndex[ch]++;
		} else if (c == 0x90) { 
			// $B2;$r$J$i$5$J$$(B
			if (ch == TRI_CHANNEL) { 
				*(char *)0x4008 = 0x00;
			} else if (ch == DMC_CHANNEL) {
				*(char *)0x4015 = 0x0f;
			} else {
				unsigned char t;
				t = ch << 2;
				*((char *)0x4000 + t) = 0x30;
			}
			SoundCounter[ch] = SoundAddr[SoundIndex[ch]];
			SoundIndex[ch]++;
		} else if (ch != NOISE_CHANNEL && ch != DMC_CHANNEL) {
			// $B6k7A$H;03Q(B
			unsigned char t;
			unsigned char frlow,frhigh;
			// $B2;Dx(B 
			t = c & 0x0f;
			FreqNumber[ch] = FreqTable[t];
			// $B%*%/%?!<%V=hM}(B
			t = c & 0xf0;
			t = t >> 4;
			while (t-- != 0) {
				FreqNumber[ch] = FreqNumber[ch] >> 1;
			}
			t = ch << 2;
			// $B;03QGH%A%c%s%M%k$OJL=hM}(B
			if (ch == TRI_CHANNEL) { 
				*(char *)0x4008 = 0xff;
			} else {
				// $B%l%8%9%?=q$-9~$_(B
				*((char *)0x4000 + t) = Ctrl0Number[ch];
				*((char *)0x4001 + t) = Ctrl1Number[ch];
			}
			frlow = (unsigned char )(0xff & FreqNumber[ch]);
			frhigh = (unsigned char )(FreqNumber[ch] >> 8);
			*((char *)0x4002 + t) = frlow;
			*((char *)0x4003 + t) = frhigh | 0x08;

			SoundCounter[ch] = SoundAddr[SoundIndex[ch]];
			SoundIndex[ch]++;
		} else if (ch == NOISE_CHANNEL){
			// $B%N%$%:%A%c%s%M%k(B
			*(char *)0x400c = Ctrl0Number[ch];
			*(char *)0x400e = c;
			*(char *)0x400f = 0x08;
			
			SoundCounter[ch] = SoundAddr[SoundIndex[ch]];
			SoundIndex[ch]++;
		} else if (ch == DMC_CHANNEL) {
			// DMC$B%A%c%s%M%k(B
			*(char *)0x4015 = 0x0f;
			*(char *)0x4010 = c;
			*(char *)0x4011 = DELTA_INIT;
			*(char *)0x4012 = SoundAddr[SoundIndex[ch]];
			SoundIndex[ch]++;
			*(char *)0x4013 = SoundAddr[SoundIndex[ch]];
			SoundIndex[ch]++;
			*(char *)0x4015 = 0x1f;
			SoundCounter[ch] = SoundAddr[SoundIndex[ch]];
			SoundIndex[ch]++;
		}

	} 
	return ;
}
